@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -23,6 +23,7 @@
 #include <math.h>
 #include <regex.h>
 #include <pthread.h>
+#include <errno.h>
 #include <assert.h>
 
 #include "relay.h"
@@ -33,8 +34,6 @@
 #include "fnv1a.h"
 
 static pthread_t aggregatorid;
-static aggregator *aggregators = NULL;
-static aggregator *lastaggr = NULL;
 static size_t prevreceived = 0;
 static size_t prevsent = 0;
 static size_t prevdropped = 0;
@@ -52,6 +51,7 @@ aggregator_new(
 		enum _aggr_timestamp tswhen)
 {
 	aggregator *ret = malloc(sizeof(aggregator));
+	int intconn[2];
 
 	if (ret == NULL)
 		return ret;
@@ -59,11 +59,14 @@ aggregator_new(
 	assert(interval != 0);
 	assert(interval < expire);
 
-	if (aggregators == NULL) {
-		aggregators = lastaggr = ret;
-	} else {
-		lastaggr = lastaggr->next = ret;
+	if (pipe(intconn) < 0) {
+		logerr("failed to create pipe for aggregator: %s\n",
+				strerror(errno));
+		free(ret);
+		return NULL;
 	}
+	ret->disp_conn = dispatch_addconnection_aggr(intconn[0]);
+	ret->fd = intconn[1];
 
 	ret->interval = interval;
 	ret->expire = expire;
@@ -357,6 +360,7 @@ cmp_entry(const void *l, const void *r)
 static void *
 aggregator_expire(void *sub)
 {
+	aggregator *aggrs = (aggregator *)sub;
 	time_t now;
 	aggregator *s;
 	struct _aggr_bucket *b;
@@ -364,11 +368,10 @@ aggregator_expire(void *sub)
 	struct _aggr_invocations *inv;
 	struct _aggr_invocations *lastinv;
 	double *values;
-	size_t len;
+	size_t len = 0;
 	int i;
 	unsigned char j;
 	int work;
-	server *submission = (server *)sub;
 	char metric[METRIC_BUFSIZ];
 	char isempty;
 	long long int ts = 0;
@@ -376,7 +379,7 @@ aggregator_expire(void *sub)
 	while (1) {
 		work = 0;
 
-		for (s = aggregators; s != NULL; s = s->next) {
+		for (s = aggrs; s != NULL; s = s->next) {
 			/* send metrics for buckets that are completely past the
 			 * expiry time, unless we are shutting down, then send
 			 * metrics for all buckets that have completed */
@@ -409,27 +412,27 @@ aggregator_expire(void *sub)
 								}
 								switch (c->type) {
 									case SUM:
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric, b->sum, ts);
 										break;
 									case CNT:
-										snprintf(metric, sizeof(metric),
-												"%s %zd %lld\n",
+										len = snprintf(metric, sizeof(metric),
+												"%s %zu %lld\n",
 												inv->metric, b->cnt, ts);
 										break;
 									case MAX:
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric, b->max, ts);
 										break;
 									case MIN:
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric, b->min, ts);
 										break;
 									case AVG:
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric,
 												b->sum / (double)b->cnt, ts);
@@ -450,7 +453,7 @@ aggregator_expire(void *sub)
 										 * iso sorting the full array */
 										qsort(values, b->cnt,
 												sizeof(double), cmp_entry);
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric,
 												values[n - 1],
@@ -464,16 +467,21 @@ aggregator_expire(void *sub)
 										for (i = 0; i < b->cnt; i++)
 											ksum += pow(values[i] - avg, 2);
 										ksum /= (double)b->cnt;
-										snprintf(metric, sizeof(metric),
+										len = snprintf(metric, sizeof(metric),
 												"%s %f %lld\n",
 												inv->metric,
 												c->type == VAR ? ksum :
 													sqrt(ksum),
 												ts);
 									}	break;
+									default:
+									   assert(0);  /* for compiler (len) */
+								}
+								if (write(s->fd, metric, len) != len) {
+									s->dropped++;
+								} else {
+									s->sent++;
 								}
-								server_send(submission, strdup(metric), 1);
-								s->sent++;
 							}
 
 							/* move the bucket to the end, to make room for
@@ -543,7 +551,7 @@ aggregator_expire(void *sub)
 	}
 
 	/* free up value buckets */
-	for (s = aggregators; s != NULL; s = s->next) {
+	while ((s = aggrs) != NULL) {
 		while (s->computes != NULL) {
 			c = s->computes;
 
@@ -562,13 +570,15 @@ aggregator_expire(void *sub)
 					free(inv->buckets);
 
 					inv = invocation->next;
-					free(inv);
+					free(invocation);
 				}
 			}
 
 			s->computes = c->next;
 			free(c);
 		}
+		aggrs = aggrs->next;
+		free(s);
 	}
 
 	return NULL;
@@ -578,12 +588,12 @@ aggregator_expire(void *sub)
  * Returns the number of aggregators defined.
  */
 size_t
-aggregator_numaggregators(void)
+aggregator_numaggregators(aggregator *aggrs)
 {
 	size_t totaggregators = 0;
 	aggregator *a;
 
-	for (a = aggregators; a != NULL; a = a->next)
+	for (a = aggrs; a != NULL; a = a->next)
 		totaggregators++;
 
 	return totaggregators;
@@ -593,13 +603,13 @@ aggregator_numaggregators(void)
  * Returns the total number of computations defined.
  */
 size_t
-aggregator_numcomputes(void)
+aggregator_numcomputes(aggregator *aggrs)
 {
 	size_t totcomputes = 0;
 	aggregator *a;
 	struct _aggr_computes *c;
 
-	for (a = aggregators; a != NULL; a = a->next)
+	for (a = aggrs; a != NULL; a = a->next)
 		for (c = a->computes; c != NULL; c = c->next)
 			totcomputes++;
 
@@ -611,9 +621,10 @@ aggregator_numcomputes(void)
  * failed, true otherwise.
  */
 int
-aggregator_start(server *submission)
+aggregator_start(aggregator *aggrs)
 {
-	if (pthread_create(&aggregatorid, NULL, aggregator_expire, submission) != 0)
+	keep_running = 1;
+	if (pthread_create(&aggregatorid, NULL, aggregator_expire, aggrs) != 0)
 		return 0;
 
 	return 1;
@@ -633,12 +644,11 @@ aggregator_stop(void)
  * Returns an approximate number of received metrics by all aggregators.
  */
 size_t
-aggregator_get_received(void)
+aggregator_get_received(aggregator *a)
 {
 	size_t totreceived = 0;
-	aggregator *a;
 
-	for (a = aggregators; a != NULL; a = a->next)
+	for ( ; a != NULL; a = a->next)
 		totreceived += a->received;
 
 	return totreceived;
@@ -649,9 +659,9 @@ aggregator_get_received(void)
  * since the last call to this function.
  */
 inline size_t
-aggregator_get_received_sub()
+aggregator_get_received_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_received();
+	size_t d = aggregator_get_received(aggrs);
 	size_t r = d - prevreceived;
 	prevreceived += d;
 	return r;
@@ -661,12 +671,11 @@ aggregator_get_received_sub()
  * Returns an approximate number of metrics sent by all aggregators.
  */
 size_t
-aggregator_get_sent(void)
+aggregator_get_sent(aggregator *a)
 {
 	size_t totsent = 0;
-	aggregator *a;
 
-	for (a = aggregators; a != NULL; a = a->next)
+	for ( ; a != NULL; a = a->next)
 		totsent += a->sent;
 
 	return totsent;
@@ -677,9 +686,9 @@ aggregator_get_sent(void)
  * since the last call to this function.
  */
 inline size_t
-aggregator_get_sent_sub()
+aggregator_get_sent_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_sent();
+	size_t d = aggregator_get_sent(aggrs);
 	size_t r = d - prevsent;
 	prevsent += d;
 	return r;
@@ -691,12 +700,11 @@ aggregator_get_sent_sub()
  * time) or if they are too much in the future.
  */
 size_t
-aggregator_get_dropped(void)
+aggregator_get_dropped(aggregator *a)
 {
 	size_t totdropped = 0;
-	aggregator *a;
 
-	for (a = aggregators; a != NULL; a = a->next)
+	for ( ; a != NULL; a = a->next)
 		totdropped += a->dropped;
 
 	return totdropped;
@@ -707,9 +715,9 @@ aggregator_get_dropped(void)
  * since the last call to this function.
  */
 inline size_t
-aggregator_get_dropped_sub()
+aggregator_get_dropped_sub(aggregator *aggrs)
 {
-	size_t d = aggregator_get_dropped();
+	size_t d = aggregator_get_dropped(aggrs);
 	size_t r = d - prevdropped;
 	prevdropped += d;
 	return r;
