@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -35,6 +35,7 @@ static char keep_running = 1;
 int collector_interval = 60;
 static char cluster_refresh_pending = 0;
 static cluster *pending_clusters = NULL;
+static aggregator *pending_aggrs = NULL;
 
 /**
  * Collects metrics from dispatchers and servers and emits them.
@@ -61,7 +62,8 @@ collector_runner(void *s)
 	time_t nextcycle;
 	char ipbuf[32];
 	char *p;
-	size_t numaggregators = aggregator_numaggregators();
+	size_t numaggregators = 0;
+	aggregator *aggrs = NULL;
 	server *submission = (server *)s;
 	server **srvs = NULL;
 	char metric[METRIC_BUFSIZ];
@@ -74,9 +76,9 @@ collector_runner(void *s)
 	size_t (*d_ticks)(dispatcher *);
 	size_t (*d_metrics)(dispatcher *);
 	size_t (*d_blackholes)(dispatcher *);
-	size_t (*a_received)(void);
-	size_t (*a_sent)(void);
-	size_t (*a_dropped)(void);
+	size_t (*a_received)(aggregator *);
+	size_t (*a_sent)(aggregator *);
+	size_t (*a_dropped)(aggregator *);
 
 	/* prepare hostname for graphite metrics */
 	snprintf(metric, sizeof(metric), "carbon.relays.%s", relay_hostname);
@@ -125,6 +127,8 @@ collector_runner(void *s)
 			if (srvs != NULL)
 				free(srvs);
 			srvs = newservers;
+			aggrs = pending_aggrs;
+			numaggregators = aggregator_numaggregators(aggrs);
 			cluster_refresh_pending = 0;
 		}
 		assert(srvs != NULL);
@@ -147,107 +151,115 @@ collector_runner(void *s)
 			totticks += ticks = d_ticks(dispatchers[i]);
 			totmetrics += metrics = d_metrics(dispatchers[i]);
 			totblackholes += blackholes = d_blackholes(dispatchers[i]);
-			snprintf(m, sizem, "dispatcher%d.metricsReceived %zd %zd\n",
+			snprintf(m, sizem, "dispatcher%d.metricsReceived %zu %zu\n",
 					i + 1, metrics, (size_t)now);
 			send(metric);
-			snprintf(m, sizem, "dispatcher%d.metricsBlackholed %zd %zd\n",
+			snprintf(m, sizem, "dispatcher%d.metricsBlackholed %zu %zu\n",
 					i + 1, blackholes, (size_t)now);
 			send(metric);
-			snprintf(m, sizem, "dispatcher%d.wallTime_us %zd %zd\n",
+			snprintf(m, sizem, "dispatcher%d.wallTime_us %zu %zu\n",
 					i + 1, ticks, (size_t)now);
 			send(metric);
 		}
-		snprintf(m, sizem, "metricsReceived %zd %zd\n",
+		snprintf(m, sizem, "metricsReceived %zu %zu\n",
 				totmetrics, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "metricsBlackholed %zd %zd\n",
+		snprintf(m, sizem, "metricsBlackholed %zu %zu\n",
 				totblackholes, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "dispatch_wallTime_us %zd %zd\n",
+		snprintf(m, sizem, "dispatch_wallTime_us %zu %zu\n",
 				totticks, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "dispatch_busy %zd %zd\n",
+		snprintf(m, sizem, "dispatch_busy %zu %zu\n",
 				dispatchers_busy, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "dispatch_idle %zd %zd\n",
+		snprintf(m, sizem, "dispatch_idle %zu %zu\n",
 				dispatchers_idle, (size_t)now);
 		send(metric);
 
+#define send_server_metrics(ipbuf, ticks, metrics, queued, stalls, dropped) \
+			snprintf(m, sizem, "destinations.%s.sent %zu %zu\n", \
+					ipbuf, metrics, (size_t)now); \
+			send(metric); \
+			snprintf(m, sizem, "destinations.%s.queued %zu %zu\n", \
+					ipbuf, queued, (size_t)now); \
+			send(metric); \
+			snprintf(m, sizem, "destinations.%s.stalls %zu %zu\n", \
+					ipbuf, stalls, (size_t)now); \
+			send(metric); \
+			snprintf(m, sizem, "destinations.%s.dropped %zu %zu\n", \
+					ipbuf, dropped, (size_t)now); \
+			send(metric); \
+			snprintf(m, sizem, "destinations.%s.wallTime_us %zu %zu\n", \
+					ipbuf, ticks, (size_t)now); \
+			send(metric);
+
 		totticks = 0;
 		totmetrics = 0;
 		totqueued = 0;
 		totstalls = 0;
 		totdropped = 0;
-		for (i = 0; srvs[i] != NULL; i++) {
-			if (server_ctype(srvs[i]) == CON_PIPE) {
-				strncpy(ipbuf, "internal", sizeof(ipbuf));
 
-				ticks = s_ticks(srvs[i]);
-				metrics = s_metrics(srvs[i]);
-				queued = server_get_queue_len(srvs[i]);
-				stalls = s_stalls(srvs[i]);
-				dropped = s_dropped(srvs[i]);
+		/* exclude internal_submission metrics from the totals to avoid
+		 * artificial doubles due to internal routing details */
+		strncpy(ipbuf, "internal", sizeof(ipbuf));
+		ticks = s_ticks(submission);
+		metrics = s_metrics(submission);
+		queued = server_get_queue_len(submission);
+		stalls = s_stalls(submission);
+		dropped = s_dropped(submission);
+		send_server_metrics(ipbuf, ticks, metrics, queued, stalls, dropped);
+
+		for (i = 0; srvs[i] != NULL; i++) {
+			if (server_ctype(srvs[i]) == CON_FILE) {
+				strncpy(ipbuf, server_ip(srvs[i]), sizeof(ipbuf));
 			} else {
 				snprintf(ipbuf, sizeof(ipbuf), "%s:%u",
 						server_ip(srvs[i]), server_port(srvs[i]));
-				for (p = ipbuf; *p != '\0'; p++)
-					if (*p == '.')
-						*p = '_';
-
-				totticks += ticks = s_ticks(srvs[i]);
-				totmetrics += metrics = s_metrics(srvs[i]);
-				totqueued += queued = server_get_queue_len(srvs[i]);
-				totstalls += stalls = s_stalls(srvs[i]);
-				totdropped += dropped = s_dropped(srvs[i]);
 			}
-			snprintf(m, sizem, "destinations.%s.sent %zd %zd\n",
-					ipbuf, metrics, (size_t)now);
-			send(metric);
-			snprintf(m, sizem, "destinations.%s.queued %zd %zd\n",
-					ipbuf, queued, (size_t)now);
-			send(metric);
-			snprintf(m, sizem, "destinations.%s.stalls %zd %zd\n",
-					ipbuf, stalls, (size_t)now);
-			send(metric);
-			snprintf(m, sizem, "destinations.%s.dropped %zd %zd\n",
-					ipbuf, dropped, (size_t)now);
-			send(metric);
-			snprintf(m, sizem, "destinations.%s.wallTime_us %zd %zd\n",
-					ipbuf, ticks, (size_t)now);
-			send(metric);
+			for (p = ipbuf; *p != '\0'; p++)
+				if (*p == '.')
+					*p = '_';
+
+			totticks += ticks = s_ticks(srvs[i]);
+			totmetrics += metrics = s_metrics(srvs[i]);
+			totqueued += queued = server_get_queue_len(srvs[i]);
+			totstalls += stalls = s_stalls(srvs[i]);
+			totdropped += dropped = s_dropped(srvs[i]);
+			send_server_metrics(ipbuf, ticks, metrics, queued, stalls, dropped);
 		}
 
-		snprintf(m, sizem, "metricsSent %zd %zd\n",
+		snprintf(m, sizem, "metricsSent %zu %zu\n",
 				totmetrics, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "metricsQueued %zd %zd\n",
+		snprintf(m, sizem, "metricsQueued %zu %zu\n",
 				totqueued, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "metricStalls %zd %zd\n",
+		snprintf(m, sizem, "metricStalls %zu %zu\n",
 				totstalls, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "metricsDropped %zd %zd\n",
+		snprintf(m, sizem, "metricsDropped %zu %zu\n",
 				totdropped, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "server_wallTime_us %zd %zd\n",
+		snprintf(m, sizem, "server_wallTime_us %zu %zu\n",
 				totticks, (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "connections %zd %zd\n",
+		snprintf(m, sizem, "connections %zu %zu\n",
 				dispatch_get_accepted_connections(), (size_t)now);
 		send(metric);
-		snprintf(m, sizem, "disconnects %zd %zd\n",
+		snprintf(m, sizem, "disconnects %zu %zu\n",
 				dispatch_get_closed_connections(), (size_t)now);
 		send(metric);
 
 		if (numaggregators > 0) {
-			snprintf(m, sizem, "aggregators.metricsReceived %zd %zd\n",
-					a_received(), (size_t)now);
+			snprintf(m, sizem, "aggregators.metricsReceived %zu %zu\n",
+					a_received(aggrs), (size_t)now);
 			send(metric);
-			snprintf(m, sizem, "aggregators.metricsSent %zd %zd\n",
-					a_sent(), (size_t)now);
+			snprintf(m, sizem, "aggregators.metricsSent %zu %zu\n",
+					a_sent(aggrs), (size_t)now);
 			send(metric);
-			snprintf(m, sizem, "aggregators.metricsDropped %zd %zd\n",
-					a_dropped(), (size_t)now);
+			snprintf(m, sizem, "aggregators.metricsDropped %zu %zu\n",
+					a_dropped(aggrs), (size_t)now);
 			send(metric);
 		}
 
@@ -274,8 +286,11 @@ collector_writer(void *unused)
 	size_t queuesize;
 	double queueusage;
 	size_t totdropped;
-	size_t numaggregators = aggregator_numaggregators();
+	size_t lastconn = 0;
+	size_t lastdisc = 0;
+	size_t numaggregators = 0;
 	server **srvs = NULL;
+	aggregator *aggrs = NULL;
 
 	while (keep_running) {
 		if (cluster_refresh_pending) {
@@ -283,10 +298,78 @@ collector_writer(void *unused)
 			if (srvs != NULL)
 				free(srvs);
 			srvs = newservers;
+			aggrs = pending_aggrs;
+			numaggregators = aggregator_numaggregators(aggrs);
 			cluster_refresh_pending = 0;
 		}
 		assert(srvs != NULL);
 		sleep(1);
+		if (debug & 1) {
+			size_t mpsout;
+			size_t totout;
+			size_t mpsdrop;
+			size_t totdrop;
+			size_t totqueue;
+			size_t mpsin;
+			size_t totin;
+			size_t totconn;
+			size_t totdisc;
+			short widle;
+			short wbusy;
+			int j;
+			/* Solaris iostat like output:
+ metrics in     metrics out    metrics drop  queue    conns     disconn   workr
+  mps     tot    mps     tot    dps     tot    cur  cps   tot  dps   tot  id bs
+99999 9999999  99999 9999999  99999 9999999  99999  999 99999  999 99999  99 99
+			 */
+			if (i % 24 == 0)
+				printf(" metrics in     metrics out    metrics drop  queue    conns     disconn   workr\n"
+						"  mps     tot    mps     tot    dps     tot    cur  cps   tot  dps   tot  id bs\n");
+
+			mpsout = totout = 0;
+			mpsdrop = totdrop = 0;
+			totqueue = 0;
+			for (j = 0; srvs[j] != NULL; j++) {
+				mpsout += server_get_metrics_sub(srvs[j]);
+				totout += server_get_metrics(srvs[j]);
+
+				mpsdrop += server_get_dropped_sub(srvs[j]);
+				totdrop += server_get_dropped(srvs[j]);
+
+				totqueue += server_get_queue_len(srvs[j]);
+			}
+			mpsin = totin = 0;
+			widle = wbusy = 0;
+			for (j = 0; dispatchers[j] != NULL; j++) {
+				mpsin += dispatch_get_metrics_sub(dispatchers[j]);
+				totin += dispatch_get_metrics(dispatchers[j]);
+
+				if (dispatch_busy(dispatchers[j])) {
+					wbusy++;
+				} else {
+					widle++;
+				}
+			}
+			totconn = dispatch_get_accepted_connections();
+			totdisc = dispatch_get_closed_connections();
+			printf("%5zu %7zu  "   /* metrics in */
+					"%5zu %7zu  "  /* metrics out */
+					"%5zu %7zu  "  /* metrics dropped */
+					"%5zu  "        /* queue */
+					"%3zu %5zu  "  /* conns */
+					"%3zu %5zu  "  /* disconns */
+					"%2d %2d\n",   /* workers */
+					mpsin, totin,
+					mpsout, totout,
+					mpsdrop, totdrop,
+					totqueue,
+					totconn - lastconn, totconn,
+					totdisc - lastdisc, totdisc,
+					widle, wbusy
+				  );
+			lastconn = totconn;
+			lastdisc = totdisc;
+		}
 		i++;
 		if (i < collector_interval)
 			continue;
@@ -299,17 +382,17 @@ collector_writer(void *unused)
 
 			if (queueusage >= 0.75)
 				logout("warning: metrics queuing up "
-						"for %s:%u: %zd metrics (%d%% of queue size)\n",
+						"for %s:%u: %zu metrics (%d%% of queue size)\n",
 						server_ip(srvs[i]), server_port(srvs[i]), queued,
 						(int)(queueusage * 100));
 		}
 		if (totdropped - lastdropped > 0)
-			logout("warning: dropped %zd metrics\n", totdropped - lastdropped);
+			logout("warning: dropped %zu metrics\n", totdropped - lastdropped);
 		lastdropped = totdropped;
 		if (numaggregators > 0) {
-			totdropped = aggregator_get_dropped();
+			totdropped = aggregator_get_dropped(aggrs);
 			if (totdropped - lastaggrdropped > 0)
-				logout("warning: aggregator dropped %zd metrics\n",
+				logout("warning: aggregator dropped %zu metrics\n",
 						totdropped - lastaggrdropped);
 			lastaggrdropped = totdropped;
 		}
@@ -324,9 +407,10 @@ collector_writer(void *unused)
  * replacement is performed at the next cycle of the collector.
  */
 inline void
-collector_schedulereload(cluster *c)
+collector_schedulereload(cluster *c, aggregator *a)
 {
 	pending_clusters = c;
+	pending_aggrs = a;
 	cluster_refresh_pending = 1;
 }
 
@@ -344,16 +428,16 @@ collector_reloadcomplete(void)
  * Initialises and starts the collector.
  */
 void
-collector_start(dispatcher **d, cluster *c, server *submission, char cum)
+collector_start(dispatcher **d, cluster *c, aggregator *a, server *submission, char cum)
 {
 	dispatchers = d;
-	collector_schedulereload(c);
+	collector_schedulereload(c, a);
 
-	if (mode == DEBUG || mode == DEBUGTEST)
+	if (mode == DEBUG || mode == DEBUGTEST || mode == DEBUGSUBMISSION)
 		debug = 1;
 	debug |= (cum ? 0 : 2);
 
-	if (mode != SUBMISSION) {
+	if (mode != SUBMISSION && mode != DEBUGSUBMISSION) {
 		if (pthread_create(&collectorid, NULL, collector_runner, submission) != 0)
 			logerr("failed to start collector!\n");
 	} else {
