@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -445,6 +445,8 @@ ch_printhashring(ch_ring *ring, FILE *f)
 			column = 0;
 		}
 	}
+	if (column != 0)
+		fprintf(f, "\n");
 }
 
 unsigned short
@@ -477,10 +479,10 @@ ch_free(ch_ring *ring)
 	ch_ring_entry *w = NULL;
 
 	for (; ring->entries != NULL; ring->entries = ring->entries->next) {
-		server_shutdown(ring->entries->server);
-		free(ring->entries->server);
-
 		if (ring->entries->malloced) {
+			server_shutdown(ring->entries->server);
+			free(ring->entries->server);
+
 			if (deletes == NULL) {
 				w = deletes = ring->entries;
 			} else {
