@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@
 #include <fcntl.h>
 #include <string.h>
 #include <errno.h>
+#include <poll.h>
 #include <pthread.h>
 #include <sys/uio.h>
 #include <sys/time.h>
@@ -53,6 +54,7 @@ typedef struct _connection {
 	size_t destlen;
 	struct timeval lastwork;
 	char hadwork:1;
+	char isaggr:1;
 } connection;
 
 struct _dispatcher {
@@ -70,6 +72,7 @@ struct _dispatcher {
 	route *routes;
 	route *pending_routes;
 	char route_refresh_pending:1;
+	char hold:1;
 	char *allowed_chars;
 };
 
@@ -127,7 +130,7 @@ dispatch_addlistener(int sock)
 	if (c == sizeof(listeners) / sizeof(connection *)) {
 		free(newconn);
 		logerr("cannot add new listener: "
-				"no more free listener slots (max = %zd)\n",
+				"no more free listener slots (max = %zu)\n",
 				sizeof(listeners) / sizeof(connection *));
 		return 1;
 	}
@@ -193,7 +196,7 @@ dispatch_addconnection(int sock)
 				sizeof(connection) * (connectionslen + CONNGROWSZ));
 		if (newlst == NULL) {
 			logerr("cannot add new connection: "
-					"out of memory allocating more slots (max = %zd)\n",
+					"out of memory allocating more slots (max = %zu)\n",
 					connectionslen);
 
 			pthread_rwlock_unlock(&connectionslock);
@@ -234,6 +237,7 @@ dispatch_addconnection(int sock)
 	connections[c].buflen = 0;
 	connections[c].needmore = 0;
 	connections[c].noexpire = 0;
+	connections[c].isaggr = 0;
 	connections[c].destlen = 0;
 	gettimeofday(&connections[c].lastwork, NULL);
 	connections[c].hadwork = 1;  /* force first iteration before stalling */
@@ -243,6 +247,27 @@ dispatch_addconnection(int sock)
 	return c;
 }
 
+/**
+ * Adds a connection which we know is from an aggregator, so direct
+ * pipe.  This is different from normal connections that we don't want
+ * to count them, never expire them, and want to recognise them when
+ * we're doing reloads.
+ */
+int
+dispatch_addconnection_aggr(int sock)
+{
+	int conn = dispatch_addconnection(sock);
+
+	if (conn == -1)
+		return 1;
+
+	connections[conn].noexpire = 1;
+	connections[conn].isaggr = 1;
+	acceptedconnections--;
+
+	return 0;
+}
+
 /**
  * Adds a pseudo-listener for datagram (UDP) sockets, which is pseudo,
  * for in fact it adds a new connection, but makes sure that connection
@@ -501,32 +526,22 @@ dispatch_runner(void *arg)
 	self->state = SLEEPING;
 
 	if (self->type == LISTENER) {
-		fd_set fds;
-		int maxfd = -1;
-		struct timeval tv;
+		struct pollfd ufds[sizeof(listeners) / sizeof(connection *)];
 		while (self->keep_running) {
-			FD_ZERO(&fds);
-			tv.tv_sec = 0;
-			tv.tv_usec = 250 * 1000;  /* 250 ms */
 			for (c = 0; c < sizeof(listeners) / sizeof(connection *); c++) {
-				conn = listeners[c];
-				if (conn == NULL)
+				if (listeners[c] == NULL)
 					break;
-				FD_SET(conn->sock, &fds);
-				if (conn->sock > maxfd)
-					maxfd = conn->sock;
+				ufds[c].fd = listeners[c]->sock;
+				ufds[c].events = POLLIN;
 			}
-			if (select(maxfd + 1, &fds, NULL, NULL, &tv) > 0) {
-				for (c = 0; c < sizeof(listeners) / sizeof(connection *); c++) {
-					conn = listeners[c];
-					if (conn == NULL)
-						break;
-					if (FD_ISSET(conn->sock, &fds)) {
+			if (poll(ufds, c, 1000) > 0) {
+				for (--c; c >= 0; c--) {
+					if (ufds[c].revents & POLLIN) {
 						int client;
 						struct sockaddr addr;
 						socklen_t addrlen = sizeof(addr);
 
-						if ((client = accept(conn->sock, &addr, &addrlen)) < 0)
+						if ((client = accept(ufds[c].fd, &addr, &addrlen)) < 0)
 						{
 							logerr("dispatch: failed to "
 									"accept() new connection: %s\n",
@@ -549,6 +564,7 @@ dispatch_runner(void *arg)
 				self->routes = self->pending_routes;
 				self->pending_routes = NULL;
 				self->route_refresh_pending = 0;
+				self->hold = 0;
 			}
 
 			pthread_rwlock_rdlock(&connectionslock);
@@ -557,6 +573,10 @@ dispatch_runner(void *arg)
 				/* atomically try to "claim" this connection */
 				if (!__sync_bool_compare_and_swap(&(conn->takenby), 0, self->id))
 					continue;
+				if (self->hold && !conn->isaggr) {
+					conn->takenby = 0;
+					continue;
+				}
 				self->state = RUNNING;
 				work += dispatch_connection(conn, self);
 			}
@@ -591,6 +611,7 @@ dispatch_new(char id, enum conntype type, route *routes, char *allowed_chars)
 	ret->keep_running = 1;
 	ret->routes = routes;
 	ret->route_refresh_pending = 0;
+	ret->hold = 0;
 	ret->allowed_chars = allowed_chars;
 	if (pthread_create(&ret->tid, NULL, dispatch_runner, ret) != 0) {
 		free(ret);
@@ -645,6 +666,19 @@ dispatch_shutdown(dispatcher *d)
 	free(d);
 }
 
+/**
+ * Requests this dispatcher to stop processing connections.  As soon as
+ * schedulereload finishes reloading the routes, this dispatcher will
+ * un-hold and continue processing connections.
+ * Returns when the dispatcher is no longer doing work.
+ */
+
+inline void
+dispatch_hold(dispatcher *d)
+{
+	d->hold = 1;
+}
+
 /**
  * Schedules routes r to be put in place for the current routes.  The
  * replacement is performed at the next cycle of the dispatcher.
