@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -46,7 +46,7 @@ int
 bindlisten(
 		int ret_stream[], int *retlen_stream,
 		int ret_dgram[], int *retlen_dgram,
-		const char *interface, unsigned short port)
+		const char *interface, unsigned short port, unsigned int backlog)
 {
 	int sock;
 	int optval;
@@ -115,7 +115,7 @@ bindlisten(
 			}
 
 			if (resw->ai_protocol == IPPROTO_TCP) {
-				if (listen(sock, 3) < 0) {  /* backlog of 3, enough? */
+				if (listen(sock, backlog) < 0) {
 					close(sock);
 					continue;
 				}
@@ -161,7 +161,7 @@ bindlisten(
 			break;
 		}
 
-		if (listen(sock, 3) < 0) {  /* backlog of 3, enough? */
+		if (listen(sock, backlog) < 0) {
 			close(sock);
 			break;
 		}
