@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -24,7 +24,7 @@
 #include <time.h>
 #include <errno.h>
 #include <assert.h>
-#if defined(_OPENMP)
+#if defined(ENABLE_OPENMP)
 # include <omp.h>
 #endif
 
@@ -48,6 +48,8 @@ static dispatcher **workers = NULL;
 static char workercnt = 0;
 static cluster *clusters = NULL;
 static route *routes = NULL;
+static aggregator *aggrs = NULL;
+static server *internal_submission = NULL;
 static char *relay_logfile = NULL;
 static FILE *relay_stdout = NULL;
 static FILE *relay_stderr = NULL;
@@ -94,7 +96,7 @@ relaylog(enum logdst dest, const char *fmt, ...)
 	len = strftime(prefix, sizeof(prefix), "[%Y-%m-%d %H:%M:%S]", tm_now);
 
 	if (!console)
-		len += snprintf(prefix + len, sizeof(prefix) - len, " (%s)", dest == LOGOUT ? "MSG" : "ERR");
+		(void)snprintf(prefix + len, sizeof(prefix) - len, " (%s)", dest == LOGOUT ? "MSG" : "ERR");
 
 	fprintf(dst, "%s ", prefix);
 
@@ -125,10 +127,10 @@ exit_handler(int sig)
 			break;
 	}
 	if (keep_running) {
-		logout("caught %s, terminating...\n", signal);
+		logout("caught %s\n", signal);
 	} else {
 		logerr("caught %s while already shutting down, "
-				"forcing exit...\n", signal);
+				"forcing exit!\n", signal);
 		exit(1);
 	}
 	keep_running = 0;
@@ -139,8 +141,12 @@ hup_handler(int sig)
 {
 	route *newroutes;
 	cluster *newclusters;
+	aggregator *newaggrs;
 	int id;
 	FILE *newfd;
+	size_t numaggregators;
+	server **servers;
+	size_t i;
 
 	logout("caught SIGHUP...\n");
 	if (relay_stderr != stderr) {
@@ -159,37 +165,67 @@ hup_handler(int sig)
 			logout("reopening logfile\n");
 		}
 	}
-	logout("reloading config from '%s'...\n", config);
 
-	if (router_readconfig(&newclusters, &newroutes,
+	logout("reloading config from '%s'\n", config);
+	if (router_readconfig(&newclusters, &newroutes, &newaggrs,
 				config, queuesize, batchsize, iotimeout) == 0)
 	{
-		logerr("failed to read configuration '%s', aborting reload\n",
-				config);
+		logerr("failed to read configuration '%s', aborting reload\n", config);
 		return;
 	}
 	router_optimise(&newroutes);
 
-	logout("reloading worker");
-	for (id = 1; id < 1 + workercnt; id++)
-		dispatch_schedulereload(workers[id + 0], newroutes);
-	for (id = 1; id < 1 + workercnt; id++) {
-		while (!dispatch_reloadcomplete(workers[id + 0]))
-			usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
-		fprintf(relay_stdout, " %d", id + 1);
-		fflush(relay_stdout);
-	}
-	fprintf(relay_stdout, "\n");
-
 	logout("reloading collector\n");
-	collector_schedulereload(newclusters);
+	collector_schedulereload(newclusters, newaggrs);
 	while (!collector_reloadcomplete())
 		usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
 
+	/* During aggregator final expiry, the dispatchers should not put
+	 * new metrics, so we have to temporarily stop them processing
+	 * connections.  However, we still need to handle aggregations that
+	 * we will stop next, so this call results in the dispatchers only
+	 * dealing with connections from aggregators.  Doing this, also
+	 * means that we can reload the config atomicly, which disrupts the
+	 * service seemingly for a bit, but results in less confusing output
+	 * in the end. */
+	logout("interrupting workers\n");
+	for (id = 1; id < 1 + workercnt; id++)
+		dispatch_hold(workers[id]);
+
+	numaggregators = aggregator_numaggregators(aggrs);
+	if (numaggregators > 0) {
+		logout("expiring aggregations\n");
+		aggregator_stop();  /* frees aggrs */
+
+		/* Now the aggregator has written everything to the
+		 * dispatchers, which will hand over to the servers with proper
+		 * metric adjustments (stubs). */
+	}
+	numaggregators = aggregator_numaggregators(newaggrs);
+	if (numaggregators > 0) {
+		if (!aggregator_start(newaggrs)) {
+			logerr("failed to start aggregator, aggregations will no "
+					"longer produce output!\n");
+		}
+	}
+
+	logout("reloading workers\n");
+	for (id = 1; id < 1 + workercnt; id++)
+		dispatch_schedulereload(workers[id], newroutes); /* un-holds */
+	for (id = 1; id < 1 + workercnt; id++) {
+		while (!dispatch_reloadcomplete(workers[id + 0]))
+			usleep((100 + (rand() % 200)) * 1000);  /* 100ms - 300ms */
+	}
+
+	servers = router_getservers(clusters);
+	for (i = 0; servers[i] != NULL; i++)
+		server_stop(servers[i]);
+	free(servers);
 	router_free(clusters, routes);
 
 	routes = newroutes;
 	clusters = newclusters;
+	aggrs = newaggrs;
 
 	logout("SIGHUP handler complete\n");
 }
@@ -197,7 +233,7 @@ hup_handler(int sig)
 static int
 get_cores(void)
 {
-#if defined(_OPENMP)
+#if defined(ENABLE_OPENMP)
 	return omp_get_num_procs();
 #else
 	return 5;
@@ -227,6 +263,7 @@ do_usage(int exitcode)
 	printf("  -b  server send batch size, defaults to 2500\n");
 	printf("  -q  server queue size, defaults to 25000\n");
 	printf("  -S  statistics sending interval in seconds, defaults to 60\n");
+	printf("  -B  connection listen backlog, defaults to 3\n");
 	printf("  -T  IO timeout in milliseconds for server connections, defaults to 600\n");
 	printf("  -m  send statistics like carbon-cache.py, e.g. not cumulative\n");
 	printf("  -c  characters to allow next to [A-Za-z0-9], defaults to -_:#\n");
@@ -249,10 +286,9 @@ main(int argc, char * const argv[])
 	int dgram_socklen = sizeof(dgram_sock) / sizeof(dgram_sock[0]);
 	char id;
 	unsigned short listenport = 2003;
+	unsigned int listenbacklog = 3;
 	int ch;
 	size_t numaggregators;
-	size_t numcomputes;
-	server *internal_submission;
 	char *listeninterface = NULL;
 	server **servers;
 	char *allowed_chars = NULL;
@@ -262,7 +298,7 @@ main(int argc, char * const argv[])
 	if (gethostname(relay_hostname, sizeof(relay_hostname)) < 0)
 		snprintf(relay_hostname, sizeof(relay_hostname), "127.0.0.1");
 
-	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:S:T:c:H:")) != -1) {
+	while ((ch = getopt(argc, argv, ":hvdmstf:i:l:p:w:b:q:S:T:c:H:B:")) != -1) {
 		switch (ch) {
 			case 'v':
 				do_version();
@@ -270,6 +306,8 @@ main(int argc, char * const argv[])
 			case 'd':
 				if (mode == TEST) {
 					mode = DEBUGTEST;
+				} else if (mode == SUBMISSION) {
+					mode = DEBUGSUBMISSION;
 				} else {
 					mode = DEBUG;
 				}
@@ -278,7 +316,11 @@ main(int argc, char * const argv[])
 				smode = SUB;
 				break;
 			case 's':
-				mode = SUBMISSION;
+				if (mode == DEBUG) {
+					mode = DEBUGSUBMISSION;
+				} else {
+					mode = SUBMISSION;
+				}
 				break;
 			case 't':
 				if (mode == DEBUG) {
@@ -349,6 +391,14 @@ main(int argc, char * const argv[])
 			case 'H':
 				snprintf(relay_hostname, sizeof(relay_hostname), "%s", optarg);
 				break;
+			case 'B': {
+				int val = atoi(optarg);
+				if (val <= 0) {
+					fprintf(stderr, "error: backlog needs to be a number >0\n");
+					do_usage(1);
+				}
+				listenbacklog = (unsigned int)val;
+			}	break;
 			case '?':
 			case ':':
 				do_usage(1);
@@ -367,7 +417,7 @@ main(int argc, char * const argv[])
 	srand(time(NULL));
 
 	if (workercnt == 0)
-		workercnt = mode == SUBMISSION ? 2 : get_cores();
+		workercnt = (mode == SUBMISSION || mode == DEBUGSUBMISSION) ? 2 : get_cores();
 
 	/* any_of failover maths need batchsize to be smaller than queuesize */
 	if (batchsize > queuesize) {
@@ -402,19 +452,20 @@ main(int argc, char * const argv[])
 	fprintf(relay_stdout, "    server queue size = %d\n", queuesize);
 	fprintf(relay_stdout, "    statistics submission interval = %ds\n",
 			collector_interval);
+	fprintf(relay_stdout, "    listen backlog = %u\n", listenbacklog);
 	fprintf(relay_stdout, "    server connection IO timeout = %dms\n",
 			iotimeout);
 	if (allowed_chars != NULL)
 		fprintf(relay_stdout, "    extra allowed characters = %s\n",
 				allowed_chars);
-	if (mode == DEBUG || mode == DEBUGTEST)
+	if (mode == DEBUG || mode == DEBUGTEST || mode == DEBUGSUBMISSION)
 		fprintf(relay_stdout, "    debug = true\n");
-	else if (mode == SUBMISSION)
+	else if (mode == SUBMISSION || mode == DEBUGSUBMISSION)
 		fprintf(relay_stdout, "    submission = true\n");
 	fprintf(relay_stdout, "    routes configuration = %s\n", config);
 	fprintf(relay_stdout, "\n");
 
-	if (router_readconfig(&clusters, &routes,
+	if (router_readconfig(&clusters, &routes, &aggrs,
 				config, queuesize, batchsize, iotimeout) == 0)
 	{
 		logerr("failed to read configuration '%s'\n", config);
@@ -422,13 +473,13 @@ main(int argc, char * const argv[])
 	}
 	router_optimise(&routes);
 
-	numaggregators = aggregator_numaggregators();
-	numcomputes = aggregator_numcomputes();
+	numaggregators = aggregator_numaggregators(aggrs);
 #define dbg (mode == DEBUG || mode == DEBUGTEST ? 2 : 0)
 	if (numaggregators > 10 && !dbg) {
 		fprintf(relay_stdout, "parsed configuration follows:\n"
-				"(%zd aggregations with %zd computations omitted "
-				"for brevity)\n", numaggregators, numcomputes);
+				"(%zu aggregations with %zu computations omitted "
+				"for brevity)\n",
+				numaggregators, aggregator_numcomputes(aggrs));
 		router_printconfig(relay_stdout, 0, clusters, routes);
 	} else {
 		fprintf(relay_stdout, "parsed configuration follows:\n");
@@ -480,7 +531,7 @@ main(int argc, char * const argv[])
 
 	if (bindlisten(stream_sock, &stream_socklen,
 				dgram_sock, &dgram_socklen,
-				listeninterface, listenport) < 0) {
+				listeninterface, listenport, listenbacklog) < 0) {
 		logerr("failed to bind on port %s:%d: %s\n",
 				listeninterface == NULL ? "" : listeninterface,
 				listenport, strerror(errno));
@@ -518,10 +569,9 @@ main(int argc, char * const argv[])
 	}
 
 	/* server used for delivering metrics produced inside the relay,
-	 * that is collector (statistics) and aggregator (aggregations) */
+	 * that is, the collector (statistics) */
 	if ((internal_submission = server_new("internal", listenport, CON_PIPE,
-					NULL, 3000 + (numcomputes * 3),
-					batchsize, iotimeout)) == NULL)
+					NULL, 3000, batchsize, iotimeout)) == NULL)
 	{
 		logerr("failed to create internal submission queue, shutting down\n");
 		keep_running = 0;
@@ -529,14 +579,15 @@ main(int argc, char * const argv[])
 
 	if (numaggregators > 0) {
 		logout("starting aggregator\n");
-		if (!aggregator_start(internal_submission)) {
+		if (!aggregator_start(aggrs)) {
 			logerr("shutting down due to failure to start aggregator\n");
 			keep_running = 0;
 		}
 	}
 
 	logout("starting statistics collector\n");
-	collector_start(&workers[1], clusters, internal_submission, smode == CUM);
+	collector_start(&workers[1], clusters, aggrs, internal_submission,
+			smode == CUM);
 
 	logout("startup sequence complete\n");
 
@@ -549,16 +600,16 @@ main(int argc, char * const argv[])
 	for (ch = 0; ch < stream_socklen; ch++)
 		dispatch_removelistener(stream_sock[ch]);
 	destroy_usock(listenport);
-	logout("listeners for port %u closed\n", listenport);
+	logout("closed listeners for port %u\n", listenport);
 	/* since workers will be freed, stop querying the structures */
 	collector_stop();
-	logout("collector stopped\n");
-	if (numaggregators > 0) {
-		aggregator_stop();
-		logout("aggregator stopped\n");
-	}
 	server_shutdown(internal_submission);
 	free(internal_submission);
+	logout("stopped collector\n");
+	if (numaggregators > 0) {
+		aggregator_stop();
+		logout("stopped aggregator\n");
+	}
 	/* give a little time for whatever the collector/aggregator wrote,
 	 * to be delivered by the dispatchers */
 	usleep(500 * 1000);  /* 500ms */
@@ -572,6 +623,7 @@ main(int argc, char * const argv[])
 		fflush(relay_stdout);
 	}
 	fprintf(relay_stdout, "\n");
+	fflush(relay_stdout);
 	free(workers);
 
 	router_shutdown();
@@ -580,9 +632,9 @@ main(int argc, char * const argv[])
 		server_stop(servers[i]);
 	free(servers);
 	router_free(clusters, routes);
-	logout("servers stopped\n");
+	logout("stopped servers\n");
 
-	logout("carbon-c-relay v%s (%s) stopped\n", VERSION, GIT_VERSION);
+	logout("stopped carbon-c-relay v%s (%s)\n", VERSION, GIT_VERSION);
 
 	return 0;
 }
