@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,11 +18,11 @@
 #ifndef HAVE_RELAY_H
 #define HAVE_RELAY_H 1
 
-#define VERSION "1.3"
+#define VERSION "1.7"
 
 #define METRIC_BUFSIZ 8192
 
-enum rmode { NORMAL, DEBUG, SUBMISSION, TEST, DEBUGTEST };
+enum rmode { NORMAL, DEBUG, SUBMISSION, DEBUGSUBMISSION, TEST, DEBUGTEST };
 
 typedef enum { CON_TCP, CON_UDP, CON_PIPE, CON_FILE } serv_ctype;
 
