@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -223,7 +223,7 @@ determine_if_regex(route *r, char *pat, int flags)
  *         (ip:port[=instance] [proto (tcp | udp)] ...)
  *     ;
  * cluster (name)
- *     log [ip]
+ *     file [ip]
  *         (/path/to/file ...)
  *     ;
  * match
@@ -262,7 +262,7 @@ determine_if_regex(route *r, char *pat, int flags)
  *    stop;
  */
 int
-router_readconfig(cluster **clret, route **rret,
+router_readconfig(cluster **clret, route **rret, aggregator **aret,
 		const char *path, size_t queuesize, size_t batchsize,
 		unsigned short iotimeout)
 {
@@ -275,6 +275,8 @@ router_readconfig(cluster **clret, route **rret,
 	struct stat st;
 	route *r = NULL;
 	route *topr = NULL;
+	aggregator *a = NULL;
+	aggregator *topa = NULL;
 	struct addrinfo *saddrs;
 	char matchcatchallfound = 0;
 
@@ -734,8 +736,8 @@ router_readconfig(cluster **clret, route **rret,
 				for (w = cl->members.ch->servers; w != NULL; w = w->next)
 					i++;
 				if (i < cl->members.ch->repl_factor) {
-					logerr("invalid cluster '%s': replication count (%zd) is "
-							"larger than the number of servers (%zd)\n",
+					logerr("invalid cluster '%s': replication count (%zu) is "
+							"larger than the number of servers (%zu)\n",
 							name, cl->members.ch->repl_factor, i);
 					free(cl);
 					free(buf);
@@ -912,7 +914,7 @@ router_readconfig(cluster **clret, route **rret,
 
 			if (matchcatchallfound) {
 				logerr("warning: match %s will never be matched "
-						"due to preceeding match * ... stop\n",
+						"due to preceding match * ... stop\n",
 						r->pattern == NULL ? "*" : r->pattern);
 			}
 			if (r->matchtype == MATCHALL && r->stop)
@@ -1150,6 +1152,11 @@ router_readconfig(cluster **clret, route **rret,
 				free(buf);
 				return 0;
 			}
+			if (a == NULL) {
+				topa = a = w->members.aggregation;
+			} else {
+				a = a->next = w->members.aggregation;
+			}
 			do {
 				if (strncmp(p, "compute", 7) != 0 || !isspace(*(p + 7))) {
 					logerr("expected 'compute' at: %.20s\n", p);
@@ -1467,6 +1474,7 @@ router_readconfig(cluster **clret, route **rret,
 	free(buf);
 	*clret = topcl;
 	*rret = topr;
+	*aret = topa;
 	return 1;
 }
 
@@ -1755,8 +1763,6 @@ router_getservers(cluster *clusters)
 }
 
 	for (c = clusters; c != NULL; c = c->next) {
-		if (c->type == BLACKHOLE || c->type == REWRITE)
-			continue;
 		if (c->type == FORWARD ||
 				c->type == FILELOG || c->type == FILELOGIP)
 		{
@@ -1892,7 +1898,7 @@ router_printconfig(FILE *f, char mode, cluster *clusters, route *routes)
 						"<unknown>",
 						ac->metric + strlen(stubname));
 			}
-			if (r->dests->next != NULL) {
+			if (!(mode & 2) && r->dests->next != NULL) {
 				destinations *dn = r->dests->next;
 				fprintf(f, "    send to");
 				if (dn->next == NULL) {
@@ -1921,7 +1927,7 @@ router_printconfig(FILE *f, char mode, cluster *clusters, route *routes)
 			for (p = r->dests->cl->name; *p != '\0' && b > blockname; p++)
 				*b-- = *p;
 			fprintf(f, "# common pattern group '%s' "
-					"contains %zd aggregations/matches\n",
+					"contains %zu aggregations/matches\n",
 					++b, cnt);
 		} else if (r->dests->cl->type == AGGRSTUB) {
 			if (mode & 2) {
@@ -2013,6 +2019,11 @@ router_free(cluster *clusters, route *routes)
 			case JUMP_CH:
 				assert(clusters->members.ch != NULL);
 				ch_free(clusters->members.ch->ring);
+				while (clusters->members.ch->servers) {
+					s = clusters->members.ch->servers->next;
+					free(clusters->members.ch->servers);
+					clusters->members.ch->servers = s;
+				}
 				free(clusters->members.ch);
 				break;
 			case FORWARD:
@@ -2030,16 +2041,18 @@ router_free(cluster *clusters, route *routes)
 				break;
 			case ANYOF:
 			case FAILOVER:
+				/* in case of secondaries, make sure nothing references
+				 * the servers anymore */
+				for (s = clusters->members.anyof->list; s != NULL; s = s->next)
+					server_shutdown(s->server);
 				while (clusters->members.anyof->list) {
-					server_shutdown(clusters->members.anyof->list->server);
 					free(clusters->members.anyof->list->server);
 
 					s = clusters->members.anyof->list->next;
 					free(clusters->members.anyof->list);
 					clusters->members.anyof->list = s;
 				}
-				if (clusters->members.anyof->servers)
-					free(clusters->members.anyof->servers);
+				free(clusters->members.anyof->servers);
 				free(clusters->members.anyof);
 				break;
 			case GROUP:
@@ -2566,8 +2579,11 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 											metric, ac->metric);
 									break;
 								}
-								snprintf(newmetric, sizeof(newmetric),
-										"%s", ac->metric + stublen);
+								len = snprintf(newmetric, sizeof(newmetric),
+										"%s", ac->metric);
+								if (len >= sizeof(newmetric))
+										len = sizeof(newmetric) - 1;
+								newfirstspace = newmetric + len;
 							}
 
 							snprintf(percentile, sizeof(percentile),
@@ -2584,13 +2600,18 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 									w->nmatch > 0 ? ac->metric + stublen : "",
 									w->nmatch > 0 ? ")" : "",
 									newmetric + stublen);
+
+							if (mode == DEBUGTEST && d->next != NULL) {
+								gotmatch |= router_test_intern(
+										newmetric,
+										newfirstspace,
+										routes);
+							}
 						}
-						if (stublen > 0) {
-							gotmatch |= router_test_intern(
-									newmetric,
-									newfirstspace,
-									routes);
+						if (mode == DEBUGTEST) {
 							return gotmatch;
+						} else {
+							gotmatch |= 4;
 						}
 					}	break;
 					case BLACKHOLE: {
@@ -2637,6 +2658,8 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 								d->cl->type == CARBON_CH ? "carbon" :
 								d->cl->type == FNV1A_CH ? "fnv1a" :
 								"jump_fnv1a", d->cl->name);
+						if (gotmatch & 4)
+							break;
 						if (mode == DEBUGTEST) {
 							fprintf(stdout, "        hash_pos(%d)\n",
 									ch_gethashpos(d->cl->members.ch->ring,
@@ -2661,6 +2684,8 @@ router_test_intern(char *metric, char *firstspace, route *routes)
 						fprintf(stdout, "    %s(%s)\n",
 								d->cl->type == ANYOF ? "any_of" : "failover",
 								d->cl->name);
+						if (gotmatch & 4)
+							break;
 						if (d->cl->type == ANYOF) {
 							const char *p;
 							fnv1a_32(hash, p, metric, firstspace);
