@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -22,6 +22,7 @@
 #include <regex.h>
 
 #include "server.h"
+#include "aggregator.h"
 
 #define CONN_DESTS_SIZE    64
 typedef struct {
@@ -34,7 +35,7 @@ typedef struct _route route;
 
 #define RE_MAX_MATCHES     64
 
-int router_readconfig(cluster **clret, route **rret, const char *path, size_t queuesize, size_t batchsize, unsigned short iotimeout);
+int router_readconfig(cluster **clret, route **rret, aggregator **aret, const char *path, size_t queuesize, size_t batchsize, unsigned short iotimeout);
 void router_optimise(route **routes);
 size_t router_rewrite_metric(char (*newmetric)[METRIC_BUFSIZ], char **newfirstspace, const char *metric, const char *firstspace, const char *replacement, const size_t nmatch, const regmatch_t *pmatch);
 void router_printconfig(FILE *f, char mode, cluster *clusters, route *routes);
