@@ -1,5 +1,5 @@
 /*
- * Copyright 2013-2015 Fabian Groffen
+ * Copyright 2013-2016 Fabian Groffen
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -20,6 +20,7 @@
 #include <string.h>
 #include <fcntl.h>
 #include <pthread.h>
+#include <poll.h>
 #include <errno.h>
 #include <sys/time.h>
 #include <sys/types.h>
@@ -278,7 +279,7 @@ server_queuereader(void *d)
 				}
 
 				/* put socket in non-blocking mode such that we can
-				 * select() (time-out) on the connect() call */
+				 * poll() (time-out) on the connect() call */
 				args = fcntl(self->fd, F_GETFL, NULL);
 				(void) fcntl(self->fd, F_SETFL, args | O_NONBLOCK);
 				ret = connect(self->fd,
@@ -287,13 +288,10 @@ server_queuereader(void *d)
 				if (ret < 0 && errno == EINPROGRESS) {
 					/* wait for connection to succeed if the OS thinks
 					 * it can succeed */
-					fd_set fds;
-					FD_ZERO(&fds);
-					FD_SET(self->fd, &fds);
-					timeoutms = self->iotimeout + (rand() % 100);
-					timeout.tv_sec = timeoutms / 1000;
-					timeout.tv_usec = (timeoutms % 1000) * 1000;
-					ret = select(self->fd + 1, NULL, &fds, NULL, &timeout);
+					struct pollfd ufds[1];
+					ufds[0].fd = self->fd;
+					ufds[0].events = POLLIN | POLLOUT;
+					ret = poll(ufds, 1, self->iotimeout + (rand() % 100));
 					if (ret == 0) {
 						/* time limit expired */
 						if (!self->failure)
@@ -307,33 +305,17 @@ server_queuereader(void *d)
 					} else if (ret < 0) {
 						/* some select error occurred */
 						if (!self->failure)
-							logerr("failed to select() for %s:%u: %s\n",
+							logerr("failed to poll() for %s:%u: %s\n",
 									self->ip, self->port, strerror(errno));
 						close(self->fd);
 						self->fd = -1;
 						self->failure += self->failure >= FAIL_WAIT_TIME ? 0 : 1;
 						continue;
 					} else {
-						int serr = 0;
-						socklen_t serrlen = sizeof(serr);
-						if (getsockopt(self->fd, SOL_SOCKET, SO_ERROR,
-									(void *)(&serr), &serrlen) < 0)
-						{
+						if (ufds[0].revents & POLLHUP) {
 							if (!self->failure)
-								logerr("failed to getsockopt() for %s:%u: %s\n",
-										self->ip, self->port, strerror(errno));
-							close(self->fd);
-							self->fd = -1;
-							self->failure += self->failure >= FAIL_WAIT_TIME ? 0 : 1;
-							continue;
-						}
-						if (serr != 0) {
-							if (!self->failure) {
-								logerr("failed to connect() to %s:%u: %s "
-										"(after select())\n",
-										self->ip, self->port, strerror(serr));
-								dispatch_check_rlimit_and_warn();
-							}
+								logerr("failed to connect() for %s:%u: "
+										"Hangup\n", self->ip, self->port);
 							close(self->fd);
 							self->fd = -1;
 							self->failure += self->failure >= FAIL_WAIT_TIME ? 0 : 1;
@@ -375,7 +357,7 @@ server_queuereader(void *d)
 		if (len != 0 && !self->keep_running) {
 			/* be noisy during shutdown so we can track any slowing down
 			 * servers, possibly preventing us to shut down */
-			logerr("shutting down %s:%u: waiting for %zd metrics\n",
+			logerr("shutting down %s:%u: waiting for %zu metrics\n",
 					self->ip, self->port, len + queue_len(self->queue));
 		}
 
@@ -398,7 +380,7 @@ server_queuereader(void *d)
 				 * partially sent data is an error for us, since we use
 				 * blocking sockets, and hence partial sent is
 				 * indication of a failure */
-				if (self->ctype != CON_UDP)
+				if (self->ctype != CON_UDP && !self->failure)
 					logerr("failed to write() to %s:%u: %s\n",
 							self->ip, self->port,
 							(slen < 0 ? strerror(errno) : "uncomplete write"));
@@ -494,6 +476,7 @@ server_new(
 
 	if (pthread_create(&ret->tid, NULL, &server_queuereader, ret) != 0) {
 		free((char *)ret->ip);
+		queue_destroy(ret->queue);
 		free(ret);
 		return NULL;
 	}
@@ -589,6 +572,7 @@ server_shutdown(server *s)
 	size_t failures;
 	size_t inqueue;
 	int err;
+	const char *p;
 
 	if (s->tid == 0)
 		return;
@@ -612,8 +596,8 @@ server_shutdown(server *s)
 			/* loop until we all failed, or nothing is in the queues */
 		} while (failures != s->secondariescnt &&
 				inqueue != 0 &&
-				logerr("any_of cluster pending %zd metrics "
-					"(with %zd failed nodes)\n", inqueue, failures) >= -1 &&
+				logerr("any_of cluster pending %zu metrics "
+					"(with %zu failed nodes)\n", inqueue, failures) >= -1 &&
 				usleep((200 + (rand() % 100)) * 1000) <= 0);
 		/* shut down entire cluster */
 		for (i = 0; i < s->secondariescnt; i++)
@@ -632,9 +616,12 @@ server_shutdown(server *s)
 	if (s->ctype == CON_TCP) {
 		size_t qlen = queue_len(s->queue);
 		if (qlen > 0)
-			logerr("dropping %zd metrics for %s:%u\n",
+			logerr("dropping %zu metrics for %s:%u\n",
 					qlen, s->ip, s->port);
 	}
+	/* drain queue not to leak the memory consumed by pending metrics */
+	while ((p = queue_dequeue(s->queue)) != NULL)
+		free((char *)p);
 	queue_destroy(s->queue);
 	free(s->batch);
 	if (s->instance)
